/*
 :******************************************************************
 :* MODULE-DESCRIPTION
 :*
 :* TITLE         (/ error output    /)
 :*
 :* NAME          errorLib.c
 :*
 :* COPYRIGHT     (C)   SIEMENS NIXDORF INFORMATIONSSYSTEME AG  1992
 :*                     ALL RIGHTS RESERVED
 :*
 :* MEMORY-CLASS  6
 :*
 :* PROG-CONVENTS (/ <internal conventions (instance, names,
 :*                  external data, ...). None -> empty text>     /)
 :*
 :* PROCEDURE     error
 :* REMARKS       (/ output error message  /)
 :* ENTRY         EXTERNAL
 :*
 :* PROCEDURE     error1
 :* REMARKS       (/ output extended error message  /)
 :* ENTRY         EXTERNAL
 :*
 :*-----------------------------------------------------------------
 :*
 :* VERSION       001
 :*   CRDATE      1992-12-02
 :*   AUTHOR      (/ Mounji Abdelaziz    , FUNDP /)
 :*   UPDATE      (/ original issue /)
 :*
 :* END-MODULE-DESCRIPTION errorLib.c
 :******************************************************************
*/
#include <stdio.h>
#include "specTab.h"
#define ErrTypes 4
static char *tabErrType[ErrTypes] = {
	"Lexical", "Syntax", "Semantics", "Fatal"};
#define ErrMsgs 14
static char *errmsgtab[ErrMsgs] = {
  /* 0 */        "'integer', 'string' or 'undef' expected",
  /* 1 */        "'integer', 'string' or a  predefined name expected",
  /* 2 */        "'integer' or 'string' expected",
  /* 3 */        "'val' or 'ref' expected",
  /* 4 */        "a procedure does not return a value",
  /* 5 */        "already part of the library: ",
  /* 6 */        "cannot open ",
  /* 7 */        "identifier expected",
  /* 8 */        "illegal option",
  /* 9 */        "out of memory",
  /*10 */        "semicolon expected",
  /*11 */        "table overflow ",
  /*12 */        "too many arguments",
  /*13 */        "unexpected symbol"
		 };
/*
 :******************************************************************
 :* UNIT-BODY errorLib.error
 :*
 :* TITLE         (/ output error message /)
 :*
 :*
 :* PURPOSE       (/ output the type, the line and the message of an error
 :*                  identified by its error code    /)
 :*                                                      END-PURPOSE
 :* VERSION       001
 :*   CRDATE      1992-12-02
 :*   AUTHOR      (/ Mounji Abdelaziz       , FUNDP              /)
 :*   UPDATE      (/ original issue /)
 :*
 :* REALIZATION   (/ <note, reference to other proc. (optional)> /)
 :*
 :*   CONTROL    (/
 :*
 :*                                                               /)
 :*   DATA       (/
 :*                 - tabErrType[]: table of error types
 :*                 - errmsgtab[] : table of error messages
 :*                 - LineNbre    : the number of the current line
 :*                                                                /)
 :*
 :*                                                          END-DATA
 :*   ASSUMPTIONS     (/ <external conditions> /)
 :*   PERFORMANCE     (/ <measures taken or not: when, why, how> /)
 :*   VALIDATION      (/ <for which param. or other data ? How ?> /)
 :*   SYNCHRONIZATION (/ <serialization & locks: purpose, method> /)
 :*   INIT&TERM       (/ <what must already/still be available> /)
 :*   ERROR-HANDLING  (/ <reaction to errors, r.c., state of data>/)
 :*   DOCUMENTS       (/
 :*                                Preliminary report on
 :*                      Advanced Security Audit trail Analysis on uniX
 :*                             (ASAX also called SAT-X)
 :*
 :*                      Advanced Security Audit trail analysis on uniX
 :*                             (ASAX also called SAT-X)
 :*                      Implementation design of the NADF Evaluator   /)
 :*                                                  END-REALIZATION
 :* PARAMETERS
 :*
 :*   Pre    (/ - 'errType' an integer value identifying the error type
 :*             - 'errorcode' an integer value identifying the error message
 :*                                                                  /)
 :*                                                             END-Pre
 :*   Post   (/ - let type    = tabErrType[errType] be the error type
 :*                   line    = 'LineNbre'
 :*                   message = errmsgtab[errorcode]
 :*              - the following line is output to stdout:
 :*             type              line: message
 :*                                                                 /)
 :*
 :*                                                             END-Post
 :*   RETURNS  NONE
 :*
 :*   REMARKS             (/ <general notes (none -> no REMARKS)> /)
 :*                                                   END-PARAMETERS
 :*
 :* END-BODY errorLib.error
 :******************************************************************
*/

error(errType, errorcode)
int errType, errorcode;
{
    ErrorFound = TRUE;
    if (errType != (int ) fatal) {
	printf("%-20s%-9s%3d: %s\n", CurrMod_name,
				     tabErrType[errType],
				     LineNbre,
				     errmsgtab[errorcode]);
    } else {
	printf("%-9s: %s\n", tabErrType[errType],
			     errmsgtab[errorcode]);
	exit(1);
    }
}
/*
 :******************************************************************
 :* UNIT-BODY errorLib.error1
 :*
 :* TITLE         (/ output extended error message /)
 :*
 :*
 :* PURPOSE       (/ output the type, the line and the message of an error
 :*                  identified by its error code then output the erronuous
 :*                  token                                              /)
 :*                                                      END-PURPOSE
 :* VERSION       001
 :*   CRDATE      1992-12-02
 :*   AUTHOR      (/ Mounji Abdelaziz       , FUNDP              /)
 :*   UPDATE      (/ original issue /)
 :*
 :* REALIZATION   (/ <note, reference to other proc. (optional)> /)
 :*
 :*   CONTROL    (/
 :*
 :*                                                               /)
 :*   DATA       (/
 :*                 - tabErrType[]: table of error types
 :*                 - errmsgtab[] : table of error messages
 :*                 - LineNbre    : the number of the current line
 :*                                                                /)
 :*
 :*                                                          END-DATA
 :*   ASSUMPTIONS     (/ <external conditions> /)
 :*   PERFORMANCE     (/ <measures taken or not: when, why, how> /)
 :*   VALIDATION      (/ <for which param. or other data ? How ?> /)
 :*   SYNCHRONIZATION (/ <serialization & locks: purpose, method> /)
 :*   INIT&TERM       (/ <what must already/still be available> /)
 :*   ERROR-HANDLING  (/ <reaction to errors, r.c., state of data>/)
 :*   DOCUMENTS       (/
 :*                                Preliminary report on
 :*                      Advanced Security Audit trail Analysis on uniX
 :*                             (ASAX also called SAT-X)
 :*
 :*                      Advanced Security Audit trail analysis on uniX
 :*                             (ASAX also called SAT-X)
 :*                      Implementation design of the NADF Evaluator   /)
 :*                                                  END-REALIZATION
 :* PARAMETERS
 :*
 :*   Pre    (/ - 'errType' an integer value identifying the error type
 :*             - 'errorcode' an integer value identifying the error message
 :*             - 'errStr' is the erronuous token
 :*                                                                  /)
 :*                                                             END-Pre
 :*   Post   (/ - let type    = tabErrType[errType] be the error type
 :*                   line    = 'LineNbre'
 :*                   message = errmsgtab[errorcode]
 :*              - the following line is output to stdout:
 :*             type              line: message  'errStr'
 :*                                                                 /)
 :*
 :*                                                             END-Post
 :*   RETURNS  NONE
 :*
 :*   REMARKS             (/ <general notes (none -> no REMARKS)> /)
 :*                                                   END-PARAMETERS
 :*
 :* END-BODY errorLib.error1
 :******************************************************************
*/
error1(errType, errorcode, errStr)
int errType, errorcode;
char *errStr;
{

    ErrorFound = TRUE;
    if (errType != (int ) fatal) {
	printf("%-20s%-9s%3d: %s: %s\n", CurrMod_name,
					 tabErrType[errType],
					 LineNbre,
					 errmsgtab[errorcode],
					 errStr);
    } else {
	printf("%-9s: %s: %s\n", tabErrType[errType],
				 errmsgtab[errorcode],
				 errStr);

    }
}
